DrDamage = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceHook-2.1")
local L = AceLibrary("AceLocale-2.2"):new("DrDamage")
local BS = AceLibrary("Babble-Spell-2.2")
local SEA = AceLibrary("SpecialEvents-Aura-2.0")
local GT = AceLibrary("Gratuity-2.0")

local defaults = {
	--Calculation
	HitCalc = false,
	HitTarget = true,
	TargetLevel = -1,
	TargetPlayer = false,
	SpellDamage = 0,
	Healing = 0,
	ShadowDamage = 0,
	FireDamage = 0,
	NatureDamage = 0,
	FrostDamage = 0,
	ArcaneDamage = 0,
	HolyDamage = 0,
	ItemCrit = 0,
	ItemHit = 0,
	ManualDmg = false,
	ManualPlus = false,	
	--Actionbar:
	DisplayType = "Avg",
	TrueAvg = true,
	FontSize = 8,
	FontXPosition = 0,
	FontYPosition = 0,
	FontColorDmg = { r = 1, g = 1, b = 0.2 },
	FontColorHeal = { r = 0.4, g = 1.0, b = 0.3 },
	CastsLeft = false,
	--Tooltip:
	Display = true,
	DefaultColor = false,
	PlusDmg = true,
	DispCrit = true,
	DispHit = true,
	AvgHit = true,
	AvgCrit = true,
	Ticks = true,
	Total = true,
	Extra = true,
	Next = true,
	DPS = true,
	DPM = true,
	Doom = true,
	Casts = true,
}

DrDamage:RegisterDB("DrDamageDB");
DrDamage:RegisterDefaults("profile", defaults )

function DrDamage:OnInitialize()	
end

local changesDMG = {
    "CHARACTER_POINTS_CHANGED",
    "PLAYER_ALIVE",
}

function DrDamage:CHARACTER_POINTS_CHANGED()
    	DrDamage:DmgUpdateTalents()
end

function DrDamage:PLAYER_ALIVE()
    	DrDamage:DmgUpdateTalents()
end

local changesAB = {
    "ACTIONBAR_PAGE_CHANGED",
    "ACTIONBAR_HIDEGRID",
    "UPDATE_SHAPESHIFT_FORM",
    "UNIT_INVENTORY_CHANGED",
    "PLAYER_TARGET_CHANGED",
    "UPDATE_MACROS",
    "UNIT_MANA"
}

local _, englishClass = UnitClass("player")
local previousValues = {
	GetSpellBonusDamage(1),
	GetSpellBonusDamage(2),
	GetSpellBonusDamage(3),
	GetSpellBonusDamage(4),
	GetSpellBonusDamage(5),
	GetSpellBonusDamage(6),
	GetSpellBonusDamage(7),
	GetSpellBonusHealing(),
	GetCombatRatingBonus(8),
	GetSpellCritChanceFromIntellect( "player" ) + GetCombatRatingBonus(11),
}


function DrDamage:ACTIONBAR_PAGE_CHANGED()

	if not self:IsEventScheduled("UpdatingAB") then
		self:ScheduleEvent("UpdatingAB", function() DrDamage:UpdateAB( false, true ) end, 0.5, self)
	end
end

function DrDamage:ACTIONBAR_HIDEGRID()

	if not self:IsEventScheduled("UpdatingAB") and ( GetCursorInfo() == "spell" or GetCursorInfo() == "macro" )  then
		self:ScheduleEvent("UpdatingAB", function() DrDamage:UpdateAB( false, true ) end, 0.5, self)
	end
end

function DrDamage:UPDATE_SHAPESHIFT_FORM()

	if not self:IsEventScheduled("UpdatingAB") then
		self:ScheduleEvent("UpdatingAB", function() DrDamage:UpdateAB( false, true ) end, 1.0, self)
	end
end

local oldBuffed
local oldDebuffed

function DrDamage:PLAYER_TARGET_CHANGED()

	if not self:IsEventScheduled("UpdatingAB") then
	
		local buffed = false
		local debuffed = false
		
		if not oldDebuffed and not oldBuffed and not DrDamage.db.profile.HitCalc then
		
			for debuffName in SEA:DebuffIter( "target" ) do

				if BS:HasReverseTranslation( debuffName ) then
					debuffName = BS:GetReverseTranslation( debuffName )
				end	

				if DrDamage.Debuffs[debuffName] then
					debuffed = true
				end
			end

			if not debuffed then
				for buffName in SEA:BuffIter( "target" ) do

					if BS:HasReverseTranslation( buffName ) then
						buffName = BS:GetReverseTranslation( buffName )
					end

					if DrDamage.HealingBuffs[buffName] then
						buffed = true
					end
				end
			end
		end
			
		if buffed or debuffed or oldDebuffed or oldBuffed or DrDamage.db.profile.HitCalc then

			self:ScheduleEvent("UpdatingAB", self.UpdateAB, 2, self)

			if debuffed then
				oldDebuffed = true
			else
				oldDebuffed = false
			end
			
			if buffed then
				oldBuffed = true
			else
				oldBuffed = false
			end
		end
	end
end

function DrDamage:UPDATE_MACROS()

	if not self:IsEventScheduled("UpdatingAB") then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1, self)
	end
end

local playerMana = UnitManaMax( "player" )

function DrDamage:UNIT_MANA( unit )

	if not DrDamage.db.profile.CastsLeft then
		do return end
	end
	
	if unit == "player" then
		
		local update = false
		local manaD = math.abs( UnitMana( "player" ) - playerMana )
	
		if manaD and manaD >= 200 then
			update = true
			playerMana = UnitMana( "player" )
		end
		
		if update and not self:IsEventScheduled( "UpdatingAB" ) then
			self:ScheduleEvent("UpdatingAB", function() DrDamage:UpdateAB( false, false, true ) end, 1, self)
		end
	end
end

function DrDamage:UNIT_INVENTORY_CHANGED( unit )

	if unit == "player" then
	
		local update = false
		
		for i = 1, 7 do
			if GetSpellBonusDamage(i) ~= previousValues[i] then
				previousValues[i] = GetSpellBonusDamage(i)
				update = true
			end
		end
		
		if GetSpellBonusHealing() ~= previousValues[8] then
			previousValues[8] = GetSpellBonusHealing()
			update = true
		end
		
		if GetCombatRatingBonus(8) ~= previousValues[9] then
			previousValues[9] = GetCombatRatingBonus(8)
			update = true
		end
		
		if DrDamage:SpellCrit() ~= previousValues[10] then
			previousValues[10] = DrDamage:SpellCrit()
			update = true
		end		
	
		if update then

			if type( DrDamage.ItemSets ) == "function" then
				DrDamage:ItemSets()
			end
		
			DrDamage:UpdateAB()			
		end
	end
end

local updateABSpell = false
local updatingSpell

function DrDamage:PlayerAuraUpdate( dataTable )

	if self:IsEventScheduled("UpdatingAB") and not updateABSpell then
		do return end
	end
	
	local update = false
	
	for i = 1, 7 do
		if GetSpellBonusDamage(i) ~= previousValues[i] then
			previousValues[i] = GetSpellBonusDamage(i)
			update = true
		end
	end

	if GetSpellBonusHealing() ~= previousValues[8] then
		previousValues[8] = GetSpellBonusHealing()
		update = true
	end

	if GetCombatRatingBonus(8) ~= previousValues[9] then
		previousValues[9] = GetCombatRatingBonus(8)
		update = true
	end

	if DrDamage:SpellCrit() ~= previousValues[10] then
		previousValues[10] = DrDamage:SpellCrit()
		update = true
	end
	
	if update then
		DrDamage:UpdateAB()
		do return end
	end
	
	if not DrDamage.DamageBuffs then
		do return end
	end
	
	local buffName = next( dataTable )
	
	if BS:HasReverseTranslation( buffName ) then
		buffName = BS:GetReverseTranslation( buffName )
	end
	
	local updateSpell
		
	if DrDamage.DamageBuffs[buffName] then
		updateSpell = DrDamage.DamageBuffs[buffName].Spell
		update = true
	elseif DrDamage.DamageDebuffs[buffName] then
		updateSpell = DrDamage.DamageDebuffs[buffName].Spell
		update = true
	end
	
	if self:IsEventScheduled("UpdatingAB") and updateABSpell then
		if updatingSpell == updateSpell then
			updateABSpell = false
		end
	end	

	if update then
		if updateSpell and not updateABSpell then
			self:ScheduleEvent("UpdatingAB", function() DrDamage:UpdateAB( false, false, false, updateSpell ) end, 0.5, self)
			updateABSpell = true
			updatingSpell = updateSpell
		else
			self:ScheduleEvent("UpdatingAB", self.UpdateAB, 0.5, self)
			updateABSpell = false
		end
	end
end


function DrDamage:TargetAuraUpdate( unitID, buffName )

	if not DrDamage.HealingBuffs or not DrDamage.Debuffs or self:IsEventScheduled("UpdatingAB") and not updateABSpell then
		do return end
	end
	
	if unitID == "target" then
	
		local update = false
		local updateSpell
		local healer = false
		
		if englishClass == "PRIEST" or englishClass == "SHAMAN" or englishClass == "PALADIN" or englishClass == "DRUID" then
			healer = true
		end
		
		if BS:HasReverseTranslation( buffName ) then
			buffName = BS:GetReverseTranslation( buffName )
		end		
		
		--|| Update healing auras only for healers ||
		if healer then
			if DrDamage.HealingBuffs[buffName] then

				if DrDamage.HealingBuffs[buffName].Class then
					for i, class in ipairs( DrDamage.HealingBuffs[buffName].Class ) do
						if class == englishClass then
							update = true
						end
					end
				elseif healer then
					update = true
				end
				
				if update then
					updateSpell = DrDamage.HealingBuffs[buffName].Spell
				end
			end
		end

		if not update then
		
			if DrDamage.Debuffs[buffName] then
				
				local debuff = DrDamage.Debuffs[buffName]
				
				if debuff.Class then
					for _, class in ipairs( debuff.Class ) do
						if class == englishClass then
							update = true
						end
					end				
				elseif debuff.School == "Healing" then
					if healer then
						update = true
					end
				elseif debuff.Affliction then
					if englishClass == "WARLOCK" then
						update = true
					end
				elseif debuff.Stun then
					if englishClass == "PALADIN" then
						update = true
					end
				else
					update = true
				end
				
				if update then
					updateSpell = DrDamage.Debuffs[buffName].Spell
				end
			end
		end
		
		if self:IsEventScheduled("UpdatingAB") and updateABSpell then
			if updatingSpell == updateSpell then
				updateABSpell = false
			end
		end
		
		if update then
			if updateSpell and not updateABSpell then
				self:ScheduleEvent("UpdatingAB", function() DrDamage:UpdateAB( false, false, false, updateSpell ) end, 0.5, self )
				updateABSpell = true
				updatingSpell = updateSpell				
			else
				self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1.5, self)
				updateABSpell = false
			end
		end
	end
end

local talents = {}
local spellInfo = {}
local talentEffects = {}
local _G

function DrDamage:OnEnable()

	if not DrDamage:DetermineClass() then
		do return end
	end
	
	talents = {}
	_G = getfenv(0)
	
	DrDamage:Hook( GameTooltip, "SetAction", true)
	DrDamage:SecureHook( GameTooltip, "SetSpell" )
	DrDamage:DmgUpdateTalents()
	
	if type( DrDamage.ItemSets ) == "function" then
		DrDamage:ItemSets()
	end
		
    	for i = 1, #changesDMG do
    		
    		if changesDMG[i] then
    			self:RegisterEvent( changesDMG[i] )
    		end
  	end
	
	if DrDamage.db.profile.DisplayType ~= "Off" then
    		
    		for i = 1, #changesAB do
    	
    			if changesAB[i] then
    				self:RegisterEvent( changesAB[i] )
    			end
    		end
    		
    		self:RegisterBucketEvent( { "SpecialEvents_PlayerBuffGained", "SpecialEvents_PlayerBuffCountChanged", "SpecialEvents_PlayerBuffLost", "SpecialEvents_PlayerDebuffGained", "SpecialEvents_PlayerDebuffCountChanged", "SpecialEvents_PlayerDebuffLost" }, 0, "PlayerAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_UnitBuffGained", "TargetAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_UnitBuffCountChanged", "TargetAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_UnitBuffLost", "TargetAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_UnitDebuffGained", "TargetAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_UnitDebuffCountChanged", "TargetAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_UnitDebuffLost", "TargetAuraUpdate" ) 
  	end  	
end

function DrDamage:OnDisable()

	DrDamage:UpdateAB( true )
	talents = nil
	spellInfo = nil
	talentEffects = nil
	_G = nil
	
end

function DrDamage:dGet (n)
	return function () return DrDamage.db.profile[n] 
	end
end
function DrDamage:dSet (n, v)
	return function (v) DrDamage.db.profile[n] = v; DrDamage:UpdateAB( false, true ); 
	end
end

local DisplayTypes = { "Avg", "DPS", "DPSC", "Off", "Min", "Max" }
local visualChange = true


--LOCALE:
local options = { 
	type='group',
	args = {
		ActionBar = {
			type = "group", desc = "Options for the actionbar", name="ActionBar",
			args = {
				DisplayType = {
					name = "Display",
					type = "text",
					desc = "Choose what to display on the actionbar",
					validate = DisplayTypes,
					get =  DrDamage:dGet("DisplayType"),
					set =  DrDamage:dSet("DisplayType", value),
					order = 70,
				},
				TrueAvg = {
					type = 'toggle',
					name = "Include crits/DoT effects.",
					desc = "Toggles if additional effects are included in the average",
					order = 71,
					get =  DrDamage:dGet( "TrueAvg" ),
					set =  DrDamage:dSet( "TrueAvg", value ),				
				},
        			Update = {
            				name = "Update",
            				desc = "Forces update to actionbar",
            				type = "execute",
            				order = 75,
            				func = "UpdateAB",
    				}, 
				FontSize = {
					type = 'range',
					name = "Actionbar font size",
					min = 6,
					max = 14,
					step = 1,
					desc = "Set the font size",
					order = 73,
					get =  DrDamage:dGet("FontSize"),
					set = function( v ) 
						DrDamage.db.profile.FontSize = v
						visualChange = true
						DrDamage:UpdateAB()
					end,
				},
				FontXPosition = {
					type = 'range',
					name = "Actionbar font horizontal position",
					min = -5,
					max = 5,
					step = 1,
					desc = "Set the font horizontal position",
					order = 74,
					get =  DrDamage:dGet("FontXPosition"),
					set = function( v ) 
						DrDamage.db.profile.FontXPosition = v
						visualChange = true
						DrDamage:UpdateAB()
					end,
				},
				FontYPosition = {
					type = 'range',
					name = "Actionbar font vertical position",
					min = -15,
					max = 30,
					step = 1,
					desc = "Set the font vertical position",
					order = 75,
					get =  DrDamage:dGet("FontYPosition"),
					set = function( v ) 
						DrDamage.db.profile.FontYPosition = v
						visualChange = true
						DrDamage:UpdateAB()
					end,
				},
				FontColorDmg = {
					type = "text",
					name = "Damage text color",
					desc = "Set the actionbar damage text color",
					usage = "<r=n, g=n, b=n> where n is 0-1",
					order = 76,
					get = function()
						return "r="..DrDamage.db.profile.FontColorDmg.r..", g="..DrDamage.db.profile.FontColorDmg.g..", b="..DrDamage.db.profile.FontColorDmg.b
					end,
					set = function( colors )
					
						for k, v in string.gmatch( string.lower( colors ), "([r|g|b])=([%d\.]+)" ) do
						
							local error = false
						
							if v and ( tonumber( v ) > 1 or tonumber( v ) < 0  ) or not v then
								error = true
							end
							
							if not error then
								if k == "r" then
									DrDamage.db.profile.FontColorDmg.r = v
								elseif k == "g" then
									DrDamage.db.profile.FontColorDmg.g = v
								elseif k == "b" then
									DrDamage.db.profile.FontColorDmg.b = v
								end
							end
						end
						
						visualChange = true
						DrDamage:UpdateAB()
					end
				},
				FontColorHeal = {
					type = "text",
					name = "Heal text color",
					desc = "Set the actionbar heal text color",
					usage = "<r=n, g=n, b=n> where n is 0-1",
					order = 77,
					get = function()
						return "r="..DrDamage.db.profile.FontColorHeal.r..", g="..DrDamage.db.profile.FontColorHeal.g..", b="..DrDamage.db.profile.FontColorHeal.b
					end,
					set = function( colors )
					
						for k, v in string.gmatch( string.lower( colors ), "([r|g|b])=([%d\.]+)" ) do
						
							local error = false
						
							if v and ( tonumber( v ) > 1 or tonumber( v ) < 0  ) or not v then
								error = true
							end
							
							if not error then
								if k == "r" then
									DrDamage.db.profile.FontColorHeal.r = v
								elseif k == "g" then
									DrDamage.db.profile.FontColorHeal.g = v
								elseif k == "b" then
									DrDamage.db.profile.FontColorHeal.b = v
								end
							end
						end
						
						visualChange = true 
						DrDamage:UpdateAB()
					end
				},				
				CastsLeft = {
					type = 'toggle',
					name = "Healing casts left",
					desc = "Toggles amount of casts left to display on actionbar",
					order = 80,
					get = DrDamage:dGet( "CastsLeft" ),
					set = DrDamage:dSet( "CastsLeft", value ),
				},				
			},
		},
		Tooltip = {
			type = "group", desc = "Options for the tooltips", name="Tooltip",
			args = {
				Display = {
					type = 'toggle',
					name = "Tooltip on/off",
					desc = "Toggles the tooltip info on/off",
					order = 50,
					get = DrDamage:dGet( "Display" ),
					set = DrDamage:dSet( "Display", value ),
				},
				DefaultColor = {
					type = 'toggle',
					name = "Default tooltip colors",
					desc = "Toggles the default blizzard tooltip colors on/off",
					order = 51,
					get = DrDamage:dGet( "DefaultColor" ),
					set = DrDamage:dSet( "DefaultColor", value ),
				},					
				PlusDmg = {
					type = 'toggle',
					name = "Show +dmg/+healing",
					desc = "Toggles displaying of +dmg/+healing",
					order = 52,
					get = DrDamage:dGet( "PlusDmg" ),
					set = DrDamage:dSet( "PlusDmg", value ),
				},
				DispCrit = {
					type = 'toggle',
					name = "Show crit %",
					desc = "Toggles displaying of crit %",
					order = 54,
					get = DrDamage:dGet( "DispCrit" ),
					set = DrDamage:dSet( "DispCrit", value ),
				},
				DispHit = {
					type = 'toggle',
					name = "Show hit %",
					desc = "Toggles displaying of hit %.",
					order = 56,
					get = DrDamage:dGet( "DispHit" ),
					set = DrDamage:dSet( "DispHit", value ),
				},				
				AvgHit = {
					type = 'toggle',
					name = "Show avg + hit range",
					desc = "Toggles displaying of avg hit",
					order = 58,
					get = DrDamage:dGet( "AvgHit" ),
					set = DrDamage:dSet( "AvgHit", value ),
				},
				AvgCrit = {
					type = 'toggle',
					name = "Show avg + crit range",
					desc = "Toggles displaying of avg crit",
					order = 60,
					get = DrDamage:dGet( "AvgCrit" ),
					set = DrDamage:dSet( "AvgCrit", value ),
				},
				Ticks = {
					type = 'toggle',
					name = "Show per tick/hit",
					desc = "Toggles displaying of per hit/tick values",
					order = 62,
					get = DrDamage:dGet( "Ticks" ),
					set = DrDamage:dSet( "Ticks", value ),
				},
				Total = {
					type = 'toggle',
					name = "Show avg total dmg/heal",
					desc = "Toggles displaying of average in total values",
					order = 64,
					get = DrDamage:dGet( "Total" ),
					set = DrDamage:dSet( "Total", value ),
				},				
				Extra = {
					type = 'toggle',
					name = "Show extra dmg/heal (DoT, Ignite, Chained)",
					desc = "Toggles displaying of extra info",
					order = 66,
					get = DrDamage:dGet( "Extra" ),
					set = DrDamage:dSet( "Extra", value ),
				},
				Next = {
					type = 'toggle',
					name = "Show next values (+1% crit, +10 dmg)",
					desc = "Toggles displaying of +1% crit, +10 dmg values",
					order = 68,
					get = DrDamage:dGet( "Next" ),
					set = DrDamage:dSet( "Next", value ),
				},
				DPS = {
					type = 'toggle',
					name = "Show DPS/HPS",
					desc = "Toggles displaying of DPS/HPS",
					order = 70,
					get = DrDamage:dGet( "DPS" ),
					set = DrDamage:dSet( "DPS", value ),
				},
				DPM = {
					type = 'toggle',
					name = "Show DPM/HPM",
					desc = "Toggles displaying of DPM/HPM",
					order = 72,
					get = DrDamage:dGet( "DPM" ),
					set = DrDamage:dSet( "DPM", value ),
				},
				Doom = {
					type = 'toggle',
					name = "Show Damage/healing until OOM",
					desc = "Toggles displaying of damage/healing until OOM",
					order = 74,
					get = DrDamage:dGet( "Doom" ),
					set = DrDamage:dSet( "Doom", value ),
				},
				Casts = {
					type = 'toggle',
					name = "Show casts and time until OOM.",
					desc = "Toggles displaying of casts and time until OOM (regen included)",
					order = 75,
					get = DrDamage:dGet( "Casts" ),
					set = DrDamage:dSet( "Casts", value ),
				},				
			},
		},		
		Calculation = {
			type = "group", desc = "Calculation options", name="Calculation",
			args = {
				HitCalc = {			
					type = 'toggle',
					name = "+Hit calculation",
					desc = "Toggles +hit calculation effects into averages on/off",
					order = 65,
					get = DrDamage:dGet( "HitCalc" ),
					set = DrDamage:dSet( "HitCalc", value ),
				},
				HitTarget = {			
					type = 'toggle',
					name = "Hit calculation by target level",
					desc = "Toggles +hit calculation by target level.",
					order = 66,
					get = DrDamage:dGet( "HitTarget" ),
					set = DrDamage:dSet( "HitTarget", value ),
				},
				TargetLevel = {
					type = 'range',
					name = "Manual target level setting",
					desc = "Manual set of target level compared to your level. Turn off hit by target level to use this.",
					min = -10,
					max = 10,
					step = 1,
					order = 67,
					get = DrDamage:dGet( "TargetLevel" ),
					set = DrDamage:dSet( "TargetLevel", value ),
				},
				TargetPlayer = {
					type = 'toggle',
					name = "Manual target is a player.",
					desc = "Toggles if your manually set target is a player.",
					order = 68,
					get = DrDamage:dGet( "TargetPlayer" ),
					set = DrDamage:dSet( "TargetPlayer", value ),
				},				
				ManualDmg = {
					type = 'toggle',
					name = "Manual variables",
					desc = "Allows you to manually set damage/healing properties",
					order = 71,
					get = DrDamage:dGet( "ManualDmg" ),
					set = DrDamage:dSet( "ManualDmg", value ),
				},
				ManualPlus = {
					type = 'toggle',
					name = "Manual damage adds to detected",
					desc = "Allows you to set if manual values should add to detected values",
					order = 72,
					get = DrDamage:dGet( "ManualPlus" ),
					set = DrDamage:dSet( "ManualPlus", value ),
				}, 
				Healing = {
					type = 'range',
					name = "Healing",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your +healing",
					order = 79,
					get = DrDamage:dGet("Healing"),
					set = DrDamage:dSet("Healing", value),
				},				
				SpellDamage = {
					type = 'range',
					name = "Spell Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your spell damage",
					order = 80,
					get = DrDamage:dGet("SpellDamage"),
					set = DrDamage:dSet("SpellDamage", value),
				},		
				FireDamage = {
					type = 'range',
					name = "Fire Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your fire +damage",
					order = 81,
					get = DrDamage:dGet("FireDamage"),
					set = DrDamage:dSet("FireDamage", value),		
		
				},
				ShadowDamage = {
					type = 'range',
					name = "Shadow Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your shadow +damage",
					order = 82,
					get = DrDamage:dGet("ShadowDamage"),
					set = DrDamage:dSet("ShadowDamage", value),		
				},
				FrostDamage = {
					type = 'range',
					name = "Frost Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your frost +damage",
					order = 83,
					get = DrDamage:dGet("FrostDamage"),
					set = DrDamage:dSet("FrostDamage", value),		
				},
				NatureDamage = {
					type = 'range',
					name = "Nature Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your nature +damage",
					order = 84,
					get = DrDamage:dGet("NatureDamage"),
					set = DrDamage:dSet("NatureDamage", value),		
				},
				ArcaneDamage = {
					type = 'range',
					name = "Arcane Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your arcane +damage",
					order = 85,
					get = DrDamage:dGet("ArcaneDamage"),
					set = DrDamage:dSet("ArcaneDamage", value),		
				},
				HolyDamage = {
					type = 'range',
					name = "Holy Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your holy +damage",
					order = 85,
					get = DrDamage:dGet("HolyDamage"),
					set = DrDamage:dSet("HolyDamage", value),		
				},					
				ItemCrit = {
					type = 'range',
					name = "Item Crit Bonus",
					min = 0,
					max = 1,
					isPercent = true,
					desc = "Input your crit bonus from items",
					order = 88,
					get = DrDamage:dGet("ItemCrit"),
					set = DrDamage:dSet("ItemCrit", value),					
				},
				ItemHit = {
					type = 'range',
					name = "Item Hit Bonus",
					min = 0,
					max = 1,
					isPercent = true,
					desc = "Input your hit bonus from items",
					order = 89,
					get = DrDamage:dGet("ItemHit"),
					set = DrDamage:dSet("ItemHit", value),	
				},
			},
		},
	},
};

DrDamage:RegisterChatCommand( { "/drdmg", "/drdamage" }, options);

function DrDamage:ClearTable( table )
	for k in pairs( table ) do
		table[k] = nil
	end
end

function DrDamage:DmgUpdateTalents()

	DrDamage:ClearTable( talents )
	
	local numTabs = GetNumTalentTabs();
	
	for t=1, numTabs do
	
    		local numTalents = GetNumTalents(t);
    		  		
	    	for i=1, numTalents do
	    	
	    		local nameTalent, _, _, _, currRank, maxRank= GetTalentInfo(t,i);
	    		
	    		if BS:HasReverseTranslation( nameTalent ) then
	    			nameTalent = BS:GetReverseTranslation( nameTalent )
	    		end
	    		
	    	    	if currRank ~= 0 then
	    	    			    	    
	    	    		if talentEffects[nameTalent] then
	    	    			table.insert( talents, { Name = nameTalent, Rank = currRank } )
	    	    			--self:Print("Talent name: " .. nameTalent .. " Points: " .. currRank .. "/" .. maxRank )
	    	    		end
	    	    	end	    	    
	    	end
	end
	
	if not self:IsEventScheduled("UpdatingAB") then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1.5, self)
	end
end


function DrDamage:DetermineClass()

	local nomatch
	local deleted

	if englishClass == "WARLOCK" then
		if type( DrDamage.WarlockData ) == "function" then
			spellInfo, talentEffects = DrDamage:WarlockData()
		else
			deleted = true
		end
	elseif englishClass == "MAGE" then
		if type( DrDamage.MageData ) == "function" then
			spellInfo, talentEffects = DrDamage:MageData()
		else
			deleted = true
		end		
	elseif englishClass == "PRIEST" then
		if type( DrDamage.PriestData ) == "function" then
			spellInfo, talentEffects = DrDamage:PriestData()
		else
			deleted = true
		end		
	elseif englishClass == "SHAMAN" then
		if type( DrDamage.ShamanData ) == "function" then
			spellInfo, talentEffects = DrDamage:ShamanData()
		else
			deleted = true
		end	
	elseif englishClass == "DRUID" then
		if type( DrDamage.DruidData ) == "function" then
			spellInfo, talentEffects = DrDamage:DruidData()
		else
			deleted = true
		end
	elseif englishClass == "PALADIN" then
		if type( DrDamage.PaladinData ) == "function" then
			spellInfo, talentEffects = DrDamage:PaladinData()
			self:Print( "Note: Paladin functionality is limited and experimental." )
		else
			deleted = true
		end	
	else
		nomatch = true
	end
	
	if deleted then
		self:Print( "The data file for " .. englishClass:lower() .. "s has been deleted." )
		return false
	elseif nomatch then
		return false
	else
		return true	
	end
end

function findpattern(text, pattern, start)
	if text and pattern and string.find( text, pattern, start) then
		return string.sub(text, string.find(text, pattern, start))
	else
		return " "
	end
end

function DrDamage:MatchData( data, name )

	if not data then
		return false
	end

	if type( data ) == "table" then
		for _, dataName in ipairs( data ) do
			if dataName == name then
				return true
			end
		end
	else
		if data == name then
			return true
		end
	end
	
	return false
end

function DrDamage:SetSpell( frame, slot )

	if not DrDamage.db.profile.Display then
		do return end
	end
		
	GT:SetSpell( slot, BOOKTYPE_SPELL )
	local spellName, spellRank = GetSpellName( slot, BOOKTYPE_SPELL );
	
	if spellRank and string.len( spellRank ) >= 4 then
		spellRank = tonumber( findpattern( spellRank, "%d+" ))
	end
	
	if not spellRank then
		spellRank = "None"
	end
	
	DrDamage:DoTooltip( frame, spellName, spellRank )

end

function DrDamage:SetAction( frame, slot )

	self.hooks[frame].SetAction(frame, slot)
	
	if not DrDamage.db.profile.Display then
		do return end
	end

	GT:SetAction( slot )	
	local gtype, _, _ = GetActionInfo( slot )
	local spellName, spellRank = frame:GetSpell()
	
	if gtype == "spell" or gtype == "macro" then
	
		if spellName then 
		
			if spellRank and string.len( spellRank ) >= 4 then
				spellRank = tonumber( findpattern( spellRank, "%d+" ))
			else
				spellRank = "None"
			end
	
			DrDamage:DoTooltip( frame, spellName, spellRank )
		end
	end
end


local TooltipData = {}

function DrDamage:DoTooltip( frame, spellName, spellRank )

	if spellName and spellRank then
  	
		if spellInfo[spellName] then
		
			baseSpell = spellInfo[spellName][0]
			tableSpell = spellInfo[spellName][spellRank]
			
			if not tableSpell then
				tableSpell = spellInfo[spellName]["None"]
			end
			
			if tableSpell and not baseSpell.NoTooltip then
				
				local healingSpell, AvgTotal, DPS, DPSC, AvgDmg, MinDmg, MaxDmg, TooltipData = DrDamage:RawNumbers ( tableSpell, spellName )
				local DPM
				local DOOM
				local _, _, ManaCost = GT:Find(L["(%d+) Mana"], 2, 2)	
				local SOOM
				local castsOOM
	
				if ManaCost then
				
					ManaCost = tonumber( ManaCost )
					
					if ManaCost > 0 then
						
						DPM = DrDamage:Round( AvgTotal / ManaCost, 1 )
						
						if TooltipData.Illumination and TooltipData.Illumination > 0 then
							DPM = DPM * ( 1 + ( TooltipData.CritRate / 100 ) * TooltipData.Illumination * 0.2 )
						end
					
						local PlayerMana = UnitMana("player")
						local regen = 0
						
						if DrDamage.db.profile.Casts and PlayerMana then
						
							local Cooldown = TooltipData.Cooldown
							local ManaRegen = TooltipData.ManaRegen
							local CastTime = TooltipData.CastTime

							if Cooldown > 0 and Cooldown <= 10 then
								CastTime = Cooldown
							end

							local castsLeft =  math.floor( PlayerMana / ManaCost )
							local secondsCastsLeft = castsLeft * CastTime
							local regenCastsTotal = 0

							if Cooldown <= 10 then

								regen = secondsCastsLeft * ManaRegen
								local regenCasts = math.floor( regen / ManaCost )
								regenCastsTotal = regenCasts
								local h = 0
								local lowCost = false

								while regenCasts > 0 do
									h = h + 1
									local newRegen = regenCasts * CastTime * ManaRegen
									regen = regen + newRegen
									local newCasts = math.floor( newRegen / ManaCost )
									regenCastsTotal = regenCastsTotal + newCasts
									regenCasts = newCasts

									if h >= 5 then
										lowCost = true
										break
									end
								end
							end

							castsOOM = castsLeft + regenCastsTotal
							
							if TooltipData.Illumination and TooltipData.Illumination > 0 then
								castsOOM = castsOOM + math.floor( TooltipData.Illumination * 0.2 * ( TooltipData.CritRate / 100 ) * castsOOM )
							end							
							
							SOOM = DrDamage:Round( castsOOM * CastTime, 1 )
							
						
							if lowCost then
								SOOM = SOOM .. "+"
								castsOOM = castsOOM .. "+"
							end
						end
						
						if PlayerMana then
							DOOM = DrDamage:Round( DPM * ( PlayerMana + regen ) )
						end
					end
				end
				
				frame:AddLine(" ")
				
				local r, g, b = 1, 0.82745098, 0
				
				if not DrDamage.db.profile.DefaultColor then
					r, g, b = 0, 0.7, 0
				end
				
				if DrDamage.db.profile.PlusDmg then
					if healingSpell then
						frame:AddDoubleLine("+Healing: ", TooltipData.SpellDmg .. " (x" .. TooltipData.SpellDmgM .. ")", 1, 1, 1, r, g, b  )
					else
						frame:AddDoubleLine("+Dmg: ", TooltipData.SpellDmg .. " (x" .. TooltipData.SpellDmgM .. ")", 1, 1, 1, r, g, b  )
					end
				end
  				
  				if DrDamage.db.profile.DispCrit and TooltipData.CritRate > 0 then
  					frame:AddDoubleLine("Crit:", TooltipData.CritRate .. "%", 1, 1, 1, r, g, b )
  				end
  				
  				if DrDamage.db.profile.DispHit and not baseSpell.Unresistable and not healingSpell then
  					frame:AddDoubleLine("Hit:", TooltipData.HitRate .. "%", 1, 1, 1, r, g, b )
  				end  				
  				
  				if not DrDamage.db.profile.DefaultColor then
  					r, g, b = 0.8, 0.8, 0.9
  				end
  				
  				if DrDamage.db.profile.AvgHit then
  					frame:AddDoubleLine("Avg (x".. TooltipData.DmgM .. "):", AvgDmg .. " (".. MinDmg .."-".. MaxDmg ..")", 1, 1, 1, r, g, b )
  					
  					if baseSpell.Leech and TooltipData.AvgLeech and TooltipData.AvgLeech > AvgDmg then
  						frame:AddDoubleLine("Avg Heal:", TooltipData.AvgLeech, 1, 1, 1, r, g, b )
  					end
  				end
  				
  				if DrDamage.db.profile.AvgCrit and TooltipData.AvgCrit > 0 and not baseSpell.sHits then
  					frame:AddDoubleLine("Avg Crit:", TooltipData.AvgCrit .. " (".. TooltipData.MinCrit .."-".. TooltipData.MaxCrit ..")", 1, 1, 1, r, g, b )
  				end
  				
  				if DrDamage.db.profile.Ticks and TooltipData.PerHit then
  					if baseSpell.sHits then
  						frame:AddDoubleLine( "Hits:", TooltipData.Hits .. "x ~" .. TooltipData.PerHit, 1, 1, 1, r, g, b )
  						
						if baseSpell.Leech and TooltipData.PerHitHeal and TooltipData.PerHitHeal > TooltipData.PerHit then
							frame:AddDoubleLine( "Hits Heal:", TooltipData.Hits .. "x ~" .. TooltipData.PerHitHeal, 1, 1, 1, r, g, b )
						end  						
  					elseif baseSpell.sTicks then
  						frame:AddDoubleLine( "Ticks:", TooltipData.Hits .. "x ~" .. TooltipData.PerHit, 1, 1, 1, r, g, b )
  						
						if baseSpell.Leech and TooltipData.PerHitHeal and TooltipData.PerHitHeal > TooltipData.PerHit then
							frame:AddDoubleLine( "Ticks Heal:", TooltipData.Hits .. "x ~" .. TooltipData.PerHitHeal, 1, 1, 1, r, g, b )
						end  						
  					end
  				end
  				
  				if DrDamage.db.profile.Extra then 
  				
  					if TooltipData.ChainDmg > 0 then
  						frame:AddDoubleLine("Avg Chain Effect:", TooltipData.ChainDmg, 1, 1, 1, r, g, b  )
  					end
  				
  					if TooltipData.IgniteDmg > 0 then
  						frame:AddDoubleLine("Ignite:", TooltipData.IgniteDmg, 1, 1, 1, r, g, b )
  					end
  				
   					if TooltipData.DotDmg > 0 then
  						frame:AddDoubleLine("DoT:", TooltipData.DotDmg, 1, 1, 1, r, g, b )
  					end
  				end  				
  				
  				if DrDamage.db.profile.Total and AvgTotal > AvgDmg then
  					frame:AddDoubleLine("Avg Total:", AvgTotal, 1, 1, 1, r, g, b  )
  				end
  				
  				if not DrDamage.db.profile.DefaultColor then
  					r, g, b = 0.3, 0.6, 0.5
  				end
  				
  				--Credits to the author of RatingBuster (Whitetooth) for the formula!
  				
				local function GetRating( hit )
					local playerLevel = UnitLevel("player")
					local base = 14
					local rating = 0
					
					if hit then
						base = 8
					end
					
					if playerLevel <= 9 then
						rating = 0
					elseif playerLevel <= 60 then
						rating = base / ( ( playerLevel - 8 ) / 52 )
					elseif playerLevel <= 70 then
						rating = base / ( 131/41 - 3/82 * playerLevel  )
					end
					
					return DrDamage:Round( rating, 2 )
				end  				
  				  				
  				if DrDamage.db.profile.Next and not baseSpell.NoNext then

  					if TooltipData.NextCrit > 0 then
  						if healingSpell then
  							frame:AddDoubleLine("+1% Crit (+" .. GetRating() .. "):", "+" .. TooltipData.NextCrit .. " (= " .. DrDamage:Round( ( TooltipData.NextCrit * 10 ) / TooltipData.NextTenDmg, 1 ) .. " +heal)", 1, 1, 1, r, g, b )	
  						else
  							frame:AddDoubleLine("+1% Crit (+" .. GetRating() .. "):", "+" .. TooltipData.NextCrit .. " (= " .. DrDamage:Round( ( TooltipData.NextCrit * 10 ) / TooltipData.NextTenDmg, 1 ) .. " +dmg)", 1, 1, 1, r, g, b )	
  						end
  					end
  					
  					if not healingSpell and not baseSpell.Unresistable then
  						frame:AddDoubleLine( "+1% Hit (+" .. GetRating( true ) .. "):", "+" .. TooltipData.NextHit, 1, 1, 1, r, g, b )
  					end
  				
  					if healingSpell then
  						frame:AddDoubleLine( "+10 Healing:", "+" .. TooltipData.NextTenDmg, 1, 1, 1, r, g, b )
  					else
  						frame:AddDoubleLine( "+10 Dmg:", "+" .. TooltipData.NextTenDmg, 1, 1, 1, r, g, b )
  					end
  					
  				end
  				
  				if not DrDamage.db.profile.DefaultColor then
  					r, g, b = 0.8, 0.1, 0.1
  				end
  				
  				if DrDamage.db.profile.DPS and not baseSpell.NoDPS then
  				
  					if DPSC ~= DPS and not baseSpell.NoDPS then
  						if healingSpell then
  							frame:AddDoubleLine("HPS/HPSC:", DPS .. "/" .. DPSC, 1, 1, 1, r, g, b)  							
  						else
  							frame:AddDoubleLine("DPS/DPSC:", DPS .. "/" .. DPSC, 1, 1, 1, r, g, b)
      						end
      					else
      				
      						if healingSpell then
      							frame:AddDoubleLine("HPS:", DPS, 1, 1, 1, r, g, b)
      						else
      							frame:AddDoubleLine("DPS:", DPS, 1, 1, 1, r, g, b)   						
      						end
      					end
      					
					if TooltipData.DPSCD > 0 then
						
						if healingSpell then
							frame:AddDoubleLine("HPS (CD):", TooltipData.DPSCD, 1, 1, 1, r, g, b)
						else
							frame:AddDoubleLine("DPS (CD):", TooltipData.DPSCD, 1, 1, 1, r, g, b)
						end
					end  					
      				end
      				
      				if DrDamage.db.profile.DPM and DPM and not baseSpell.NoDPM then
      					if healingSpell then
      						frame:AddDoubleLine("HPM:", DPM, 1, 1, 1, r, g, b )
      					else
      						frame:AddDoubleLine("DPM:", DPM, 1, 1, 1, r, g, b )
      					end
      				end
      				
      				if DrDamage.db.profile.Doom and DOOM and not baseSpell.NoDoom then
      					if healingSpell then
      						frame:AddDoubleLine("Healing until OOM:", DOOM, 1, 1, 1, r, g, b )
      					else
      						frame:AddDoubleLine("Damage until OOM:", DOOM, 1, 1, 1, r, g, b )
      					end
      				end
      				
      				if DrDamage.db.profile.Casts and castsOOM and SOOM then
      						if TooltipData.Cooldown <= 10 and not baseSpell.eDot then
      							frame:AddDoubleLine("Casts (w/ regen):", castsOOM .." (" .. SOOM .. "s)", 1, 1, 1, r, g, b )
      						else
      							frame:AddDoubleLine("Casts:", castsOOM, 1, 1, 1, r, g, b )
      						end
      				end      				
      				
				DrDamage:ClearTable( TooltipData )
      				
      				frame:Show()			
			end  			
		end	  
	end
end
	
local ABtable
	
function DrDamage:UpdateAB( disable, hide, manaUpdate, updateSpell )

	--[[
	if updateSpell then
		self:Print( "Update AB: Spell" )
	else
		self:Print( "Update AB: All" )
	end
	--]]

	local BT3 = false
	local Bongos = false
	local DAB = false
	local NAB = false
	local Trinity = false
	local CTBar = false
	local blizzAB = false
	
	if IsAddOnLoaded("Bartender3") then
		BT3 = true
		if not ABtable then
			ABtable = Bartender3.actionbuttons
		end
	elseif IsAddOnLoaded( "Bongos" ) then
		Bongos = true
	elseif IsAddOnLoaded("DiscordActionBars") then
		DAB = true
	elseif IsAddOnLoaded("Nurfed_ActionBars") then
		NAB = true
	elseif IsAddOnLoaded("TrinityBars") then
		Trinity = true
	elseif IsAddOnLoaded("CT_BarMod") then
		CTBar = true
		if not ABtable then
			ABtable = CT_BarMod.actionButtonList
		end
	else
		blizzAB = true
		if not ABtable then
			ABtable = { "ActionButton", "BonusActionButton", "MultiBarBottomLeftButton", "MultiBarBottomRightButton","MultiBarRightButton", "MultiBarLeftButton", }
		end
	end
	
	local function ProcessButton( button, index, main )
		
		if disable or hide then
		
			for _,frame in ipairs( { button:GetChildren() } ) do

				if frame and frame.text and frame:GetID() == 200 then
					frame.text:Hide()
				end
			end
		end
   		
   		if DrDamage.db.profile.DisplayType == "Off" or disable then
   			do return end
   		end
   				
		if button:IsVisible() then
		
			local id
		
			if BT3 and ABtable[index] and ABtable[index].state == "used" then
				id = ABtable[index].object:PagedID()
			elseif Bongos then
				id = button:GetPagedID()
			elseif DAB then
				id = button:GetActionID()
			elseif NAB or Trinity then
				id = button:GetID()
			elseif CTBar then
				if main then
					id = ActionButton_GetPagedID( button )
				elseif ABtable[index] and ABtable[index].hasAction then
					id = index
				end
			elseif blizzAB then
				id = ActionButton_GetPagedID( button )
			end
			
			if id then
				DrDamage:CheckAction( id, button, manaUpdate, updateSpell )
			end
		end
	end
	
	
	if CTBar then
	
		local button
		
		for i = 1, 12 do
			button = _G["ActionButton"..i]
			
			if button then
				ProcessButton( button, i, true )
			end
		end		
		
		for key, value in pairs( ABtable ) do 
			
			button = value.button
			
			if button then
				ProcessButton( button, key )
			end
		end
		

	else
		local button
		
		for i = 1, 120 do
	
			if BT3 then
				button = _G["BT3Button"..i]
			elseif Bongos then
				button = _G["BActionButton"..i]
			elseif DAB then
				button = _G["DAB_ActionButton_"..i]
			elseif NAB then
				button = _G["Nurfed_ActionButton"..i]
			elseif Trinity then
				button = _G["TrinityActionButton"..i]
			elseif blizzAB then
				local abType = 1 + math.floor( ( i - 1 ) / 12 )
				local buttonNr = i - math.floor( ( i - 1 ) / 12 ) * 12
			
				if abType > #ABtable then
					do return end
				end
			
				button = _G[ABtable[abType]..buttonNr]
			end
					
			if button then
				ProcessButton( button, i )
			end
		end
	end
	
	if updateSpell and updateABSpell then
		updateABSpell = false
	end
	
	if updatingSpell then
		updatingSpell = nil
	end
	
	if visualChange then
		visualChange = false
	end
end

function DrDamage:CheckAction( id, button, manaUpdate, updateSpell )

	local gtype, pid = GetActionInfo(id)
	
	if gtype == "spell" or gtype =="macro" then
		
		local spellName, spellRank
		
		if gtype == "spell" then
			spellName, spellRank = GetSpellName( pid, BOOKTYPE_SPELL );
		else
			GT:SetAction( id )
			spellName = GT:GetLine(1, false)
			spellRank = GT:GetLine(1, true)
			
			if not spellName and not spellRank then
				local macroText = getglobal(button:GetName() .. "Name")

				if not macroText:IsVisible() then
					macroText:Show()
				end			
			end
		end
			
		if spellName and spellRank and spellInfo[spellName] then
		
			if updateSpell then
			
				if not DrDamage:MatchData( updateSpell, spellName ) then
					do return end
				end
				
				updateABSpell = false
			end
			
			local baseSpell = spellInfo[spellName][0]
			local tableSpell = spellInfo[spellName][ tonumber( findpattern( spellRank, "%d+" ) ) ]
						
			if not tableSpell then
				tableSpell = spellInfo[spellName]["None"]
			end
		
			if tableSpell then
			
				local healingSpell = false
				
				if baseSpell.Healing then
					healingSpell = true
				end
				
				if type( baseSpell.School ) == "table" then
					if baseSpell.School[2] == "Healing" then
						healingSpell = true
					end
				end

				if gtype == "macro" then
				
					local macroText = getglobal(button:GetName() .. "Name")

					if macroText then
						macroText:Hide()
					end

					if IsAddOnLoaded("CT_BarMod") then
						if button.name then
							button.name:SetText( "" )
						end
					end
				end
					
				local didCasts = false	
				
				if DrDamage.db.profile.CastsLeft and healingSpell then

					GT:SetAction( id )
					local _, _, ManaCost = GT:Find(L["(%d+) Mana"], 2, 2)
					local PlayerMana = UnitMana("player")

					if ManaCost then
						ManaCost = tonumber( ManaCost )
						didCasts = true
						
						if ManaCost > 0 then
							DrDamage:CreateABtext( button, math.floor( PlayerMana / ManaCost ), healingSpell )
						else
							DrDamage:CreateABtext( button, "Inf.", healingSpell )
						end
					end
				end
				
				if manaUpdate then
					do return end
				end
		
				if not didCasts then
				
					local _, AvgTotal, DPS, DPSC, Avg, MinDmg, MaxDmg = DrDamage:RawNumbers ( tableSpell, spellName )			
					local displayType = DrDamage.db.profile.DisplayType
					
					if displayType == "Avg" or displayType == "Min" or displayType == "Max" or baseSpell.NoDPS then
						if displayType == "Min" then
							DrDamage:CreateABtext( button, MinDmg, healingSpell )
						elseif displayType == "Max" then
							DrDamage:CreateABtext( button, MaxDmg, healingSpell )
						else
							if not DrDamage.db.profile.TrueAvg then
								DrDamage:CreateABtext( button, Avg, healingSpell )
							else
								DrDamage:CreateABtext( button, AvgTotal, healingSpell )
							end
						end
					elseif displayType == "DPS" then
						DrDamage:CreateABtext( button, DPS, healingSpell )
					elseif displayType == "DPSC" then
						DrDamage:CreateABtext( button, DPSC, healingSpell )
					end
				end
			end				
		end
	end
end


function DrDamage:CreateABtext( button, cDmg, HealingSpell )

	local found = false
	local positionMod = 0
	cDmg = DrDamage:Round( cDmg )
	
	if cDmg and string.len( cDmg ) then
		positionMod = math.floor( (4 - string.len( cDmg ) ) * 3.5 )
	end
	
   	for _,frame in ipairs( { button:GetChildren() } ) do
   			
   		if frame and frame.text and frame:GetID() == 200 then
   			
   			if visualChange then
				frame.text:SetPoint("BOTTOMLEFT", frame, "BOTTOM", (DrDamage.db.profile.FontXPosition - 15) + positionMod, (DrDamage.db.profile.FontYPosition + 5) )
				frame.text:SetFont( STANDARD_TEXT_FONT, DrDamage.db.profile.FontSize, "OUTLINE" )

				if HealingSpell then
					frame.text:SetTextColor( DrDamage.db.profile.FontColorHeal.r, DrDamage.db.profile.FontColorHeal.g, DrDamage.db.profile.FontColorHeal.b )
				else
					frame.text:SetTextColor( DrDamage.db.profile.FontColorDmg.r, DrDamage.db.profile.FontColorDmg.g, DrDamage.db.profile.FontColorDmg.b )
				end
			end
			
   			frame.text:SetText( cDmg )
   			frame.text:Show()
   			found = true
   		end	
   	end
   	
   	if not found then
   	
		local dpsText = CreateFrame("Frame", "DrDamageText", button )
		dpsText:SetAllPoints( button )
		dpsText:SetFrameLevel(dpsText:GetFrameLevel() + 1)
		dpsText:SetID( 200 )
		dpsText.text = dpsText:CreateFontString( nil, "OVERLAY" )
		dpsText.text:SetFontObject(GameFontNormalSmall)
		dpsText.text:SetPoint("BOTTOMLEFT", dpsText, "BOTTOM", (DrDamage.db.profile.FontXPosition - 15) + positionMod, (DrDamage.db.profile.FontYPosition + 5) )
		dpsText.text:SetFont(STANDARD_TEXT_FONT, DrDamage.db.profile.FontSize, "OUTLINE")
		
		if HealingSpell then
			dpsText.text:SetTextColor( DrDamage.db.profile.FontColorHeal.r, DrDamage.db.profile.FontColorHeal.g, DrDamage.db.profile.FontColorHeal.b )
		else
			dpsText.text:SetTextColor( DrDamage.db.profile.FontColorDmg.r, DrDamage.db.profile.FontColorDmg.g, DrDamage.db.profile.FontColorDmg.b )
		end
		
		dpsText.text:SetText( cDmg )
		dpsText.text:Show()
	end
end

function DrDamage:SpellCrit()
	return GetSpellCritChanceFromIntellect( "player" ) + GetCombatRatingBonus(11)
end


local hitDataMOB = { 95, 94, 83, 72, 61, 50, 39, 28, 17, 6 }
local hitDataPlayer = { 95, 94, 87, 80, 73, 66, 59, 52, 45, 38 }
local lastTargetLevel
local lastSpellHit

function DrDamage:SpellHit()

	local targetLevel = UnitLevel( "target" )
	local playerLevel = UnitLevel( "player" )
	local hitPerc = 99
	
	if DrDamage.db.profile.HitTarget then
		
		local levelDiff = 0

		if playerLevel >= 60 and targetLevel == -1 then
			levelDiff = 3
		elseif targetLevel == -1 then
			levelDiff = 10
		elseif targetLevel > playerLevel then
			levelDiff = targetLevel - playerLevel
		elseif targetLevel == playerLevel then
			hitPerc = 96
		end
		
		if levelDiff > 0 and levelDiff <= 10 then
		
			if UnitIsPlayer( "target" ) then
				hitPerc = hitDataPlayer[levelDiff]
			else
				hitPerc = hitDataMOB[levelDiff]
			end
		end
		
		lastTargetLevel = targetLevel
		lastSpellHit = hitPerc
		
	elseif not DrDamage.db.profile.HitTarget then
		
		if DrDamage.db.profile.TargetLevel >= 1 then
			
			if DrDamage.db.profile.TargetPlayer then
					hitPerc = hitDataPlayer[ math.floor( DrDamage.db.profile.TargetLevel ) ]
			else
					hitPerc = hitDataMOB[ math.floor( DrDamage.db.profile.TargetLevel ) ]
			end
					
		elseif math.floor( DrDamage.db.profile.TargetLevel ) == 0 then
			hitPerc = 96
		end
		
		lastTargetLevel = DrDamage.db.profile.TargetLevel
		lastSpellHit = hitPerc		
	end
		
	return hitPerc
end

function DrDamage:GetPlusDamage( school )

	local spellDMG = GetSpellBonusDamage(1)
	
	if school == "Holy" then
  		return GetSpellBonusDamage(2)
	elseif school == "Fire" then
  		return GetSpellBonusDamage(3)
  	elseif school == "Nature" then
  		return GetSpellBonusDamage(4)
  	elseif school == "Frost" then
  		return GetSpellBonusDamage(5)
  	elseif school == "Shadow" then
  		return GetSpellBonusDamage(6)  		
   	elseif school == "Arcane" then
  		return GetSpellBonusDamage(7)
  	else
  		return spellDMG
  	end	
end

local CalculationResults = {}

function DrDamage:RawNumbers( spell, spellName )

	if not spell or not spellName then
		do return end
	end
			
	local baseSpell = spellInfo[spellName][0]
	
	if BS:HasReverseTranslation( spellName ) then
		spellName = BS:GetReverseTranslation( spellName )
	end		

	local healingSpell = false
	local spellSchool 
	local spellType
	
	if type( baseSpell.School ) == "table" then
		spellSchool = baseSpell.School[1]
		spellType = baseSpell.School[2]
		
		if spellType == "Healing" then
			healingSpell = true
		end
	else
		spellSchool = baseSpell.School
	end
	
	if baseSpell.Healing then
		healingSpell = true
	end	
	
	local critPerc = DrDamage:SpellCrit() 		--Calculate crit
	local hitPerc = lastSpellHit
	local manaRegen, manaRegenCasting = GetManaRegen( "player" )
	
	--Calculate base hit chance only IF needed:
	local targetLevel = UnitLevel( "target" )
	
	if DrDamage.db.profile.HitTarget and targetLevel ~= lastTargetLevel then
		hitPerc = DrDamage:SpellHit() 		
	elseif not DrDamage.db.profile.HitTarget and DrDamage.db.profile.TargetLevel ~= lastTargetLevel then
		hitPerc = DrDamage:SpellHit()
	end
		
	hitPerc = hitPerc + GetCombatRatingBonus( 8 )   --Add Gear
	
	
	local spellDmg
	if healingSpell then
		spellDmg = GetSpellBonusHealing() 			-- +healing
	else
		spellDmg = DrDamage:GetPlusDamage( spellSchool ) 	--+all spelldamage
	end
		
	--Random item bonuses to base values not supplied by API
	
	if healingSpell then
		--something?
	else
		if UnitCreatureType( "target" ) == "Undead" then
			--Rune of the Dawn
			if IsEquippedItem( 19182 ) then
				spellDmg = spellDmg + 48
			end
		end
	end
	
	if DrDamage.db.profile.ManualDmg then
		
		if not DrDamage.db.profile.ManualPlus then
			
			if healingSpell then
				spellDmg = DrDamage.db.profile.Healing
			else
				spellDmg = DrDamage.db.profile.SpellDamage
			end
			
			critPerc = critPerc - GetCombatRatingBonus(11) + DrDamage.db.profile.ItemCrit * 100
			hitPerc = hitPerc - GetCombatRatingBonus(8) + DrDamage.db.profile.ItemHit * 100
		else
			if healingSpell then
				spellDmg = spellDmg + DrDamage.db.profile.Healing
			else
				spellDmg = spellDmg + DrDamage.db.profile.SpellDamage
			end
			
			critPerc = critPerc + DrDamage.db.profile.ItemCrit
			hitPerc = hitPerc + DrDamage.db.profile.ItemHit
		end
		
		if not healingSpell then
		
			if ( spellSchool == "Shadow" ) then
				spellDmg = spellDmg + DrDamage.db.profile.ShadowDamage --add shadowdamage
			elseif ( spellSchool == "Fire" ) then
				spellDmg = spellDmg + DrDamage.db.profile.FireDamage --add firedamage
			elseif ( spellSchool == "Nature" ) then
				spellDmg = spellDmg + DrDamage.db.profile.NatureDamage --add naturedamage
			elseif ( spellSchool == "Frost" ) then
				spellDmg = spellDmg + DrDamage.db.profile.FrostDamage --add frostdamage
			elseif ( spellSchool == "Arcane" ) then
				spellDmg = spellDmg + DrDamage.db.profile.ArcaneDamage --add arcanedamage
			elseif ( spellSchool == "Holy" ) then
				spellDmg = spellDmg + DrDamage.db.profile.HolyDamage --add holydamage			
			end
		else
			spellDmg = spellDmg + DrDamage.db.profile.Healing
		end
	end
					
	local critM = 0.5			--Crit multiplier
	local bNukeDmg = 0			--Modifier for spells with a DOT portion
	local dmgM = 1				--Damage multiplier (total)
	local dmgMAdd = 0			--Damage multiplier additive portion (for talent calculation)
	local bDmgM = 1				--Base damage Modifier
	local spellDmgM				--Spell damage multiplier
	local dotSpellDmgM = 0			--Dot portion spelldamage modifier
	local igniteM = 0			--Ignites % of total
	local cooldown = 0			--Spell's cooldown
	local finalMod = 0			--Modifier to final damage +/-
	local BuffTalentRank = 0		--Special variable for buffs with no API rank (MD for example)
	local BuffTalentRank2 = 0		--As above
	local BuffTalentRank3 = 0		--As above
	local illumination = 0			--Paladin illumination talent
	local SSRank = 0			--Special variable for Soul Siphon
	local LeechBonus = 1			--Variable for spells with healing in addition to damage
	local minDam = spell[1]			--Spell initial base min damage
	local maxDam = spell[2]			--Spell initial base max damage
	local chainFactor = 0			--Chain effect spells
	
	if baseSpell.chainFactor then
		chainFactor = baseSpell.chainFactor
	end
	
	if baseSpell.Cooldown then
		cooldown = baseSpell.Cooldown
	end

	local castTime = 1.5
	
	if spell.castTime then
		castTime = spell.castTime
	elseif baseSpell.castTime then
		castTime = baseSpell.castTime
	end
	
	local eDuration = 1
	
	if spell.eDuration then
		eDuration = spell.eDuration
	elseif baseSpell.eDuration then
		eDuration = baseSpell.eDuration
	end
	
	if baseSpell.bonusFactor then
		spellDmgM = baseSpell.bonusFactor
	elseif baseSpell.eDot then 
		spellDmgM = ( eDuration / 15)
	else
		spellDmgM = ( castTime / 3.5) 
	end
	
	local sHits
	
	if spell.sHits then
		sHits = spell.sHits
	elseif baseSpell.sHits then
		sHits = baseSpell.sHits
	end
				
	for i, charTalent in ipairs( talents ) do
		
		local talentName = charTalent.Name
		local talentRank = charTalent.Rank
		
		if talentName then
						
			if talentEffects[talentName] then
				
				local effectTalent = talentEffects[talentName]
				
				function DoCalc( ModType, talentSpell, talentValue, add, multiply )
				
					if talentSpell == "All" or talentSpell == spellSchool and not baseSpell.NoSchoolTalents or talentSpell == spellType or talentSpell == spellName then
					
						if not ModType then
						
							if add then
								dmgMAdd = dmgMAdd + talentValue
							else
								dmgM = dmgM * ( 1 + talentValue )	
							end					
						elseif ModType == "SpellDamage" then
						
							if multiply then
								spellDmgM = spellDmgM * ( 1 + talentValue )
							else
								spellDmgM = spellDmgM + talentValue
							end
						elseif ModType == "Crit" then
							critPerc = critPerc + talentValue
						elseif ModType == "Hit" then
							hitPerc = hitPerc + talentValue							
						elseif ModType == "CastTime" then
							
							if baseSpell.CastMod then
								castTime = castTime - baseSpell.CastMod * talentValue
							end
							
							if baseSpell.CastModDmg then
								spellDmgM = spellDmgM - baseSpell.CastModDmg * talentValue
							end
						elseif ModType == "BaseNuke" then
							bNukeDmg = bNukeDmg + talentValue
						elseif ModType == "BaseDamage" then
							bDmgM = bDmgM + talentValue
						elseif ModType == "DotSpellDamage" then
							dotSpellDmgM = dotSpellDmgM + talentValue
						elseif ModType == "CritMultiplier" then
							critM = critM + talentValue
						elseif ModType == "SpellDuration" then
							eDuration = eDuration + talentValue
						elseif ModType == "Cooldown" then
							cooldown = cooldown - talentValue
						elseif ModType == "Ignite" then
							igniteM = talentValue							
						elseif ModType == "BuffTalentRank" then
							BuffTalentRank = talentValue
						elseif ModType == "BuffTalentRank2" then
							BuffTalentRank2 = talentValue
						elseif ModType == "BuffTalentRank3" then
							BuffTalentRank3 = talentValue
						elseif ModType == "Illumination" then
							illumination = talentValue
						elseif ModType == "Soul Siphon" then
							SSRank = talentValue
						end
					end				
				
				end
						
				local modType = effectTalent.ModType
				local add = effectTalent.Add
				local multiply = effectTalent.Multiply
				local talentValue
				
				if type( effectTalent.Effect ) == "table" then
					talentValue = effectTalent.Effect[talentRank]
				else
					talentValue = talentRank * effectTalent.Effect
				end
				
				for j, talentSpell in ipairs( effectTalent.Spells ) do
					DoCalc( modType, talentSpell, talentValue, add, multiply  )
				end
									
				if effectTalent.Spells2 then
				
					modType = effectTalent.ModType2
					add = effectTalent.Add2
					multiply = effectTalent.Multiply2
					
					if effectTalent.Effect2 then
						talentValue = talentRank * effectTalent.Effect2
					end
					
					for j, talentSpell in ipairs( effectTalent.Spells2 ) do
						DoCalc( modType, talentSpell, talentValue, add, multiply )
					end
				end	
			end
		end		
	end
	
	--Add up additive components:
	dmgM = dmgM + dmgMAdd
	
	
	local function BuffCalc( bName, bData, bApps, bTexture, bRank, bIndex )
	
		local modType = bData.ModType
		
		if bData.Ranks then
			local buffRank

			if bRank and string.len( bRank ) >= 6 then
				buffRank = tonumber( findpattern( bRank, "%d+" ) )
			end

			if not buffRank then
				buffRank = bData.Ranks
			end

			if not modType then
				dmgM = dmgM * ( 1 + buffRank * bData.Value )
			elseif modType == "Amplify Magic" then
				spellDmg = spellDmg + bData.Value[buffRank]				
			elseif modType == "JotC" then
				local bonus = bData.Value[buffRank]

				if BuffTalentRank2 > 0 then
					bonus = bonus * ( 1 + 0.05 * BuffTalentRank2 )
				end

				if BuffTalentRank3 > 0 then
					critPerc = critPerc + BuffTalentRank3
				end							

				if bonus then
					spellDmg = spellDmg + bonus
				end
			end
		elseif bData.Apps then
			if modType == "Crit" and bApps then
				critPerc = critPerc + bApps * bData.Value
			elseif not modType and bApps then
				if GetRealZoneText() == "Naxxramas" and ( bName == "Positive Charge" or bName == "Negative Charge" ) then
					dmgM = dmgM * ( 1 + bApps * ( bData.Value / 10 ) )
				else
					dmgM = dmgM * ( 1 + bApps * bData.Value )
				end			
			end
		elseif bData.Texture and bTexture and string.find( bTexture, bData.Texture ) then
			dmgM = dmgM * ( 1 + bData.Value )
		elseif bData.Apps2 then
			dmgM = dmgM * ( 1 + bApps * bData.Value2 )
		elseif bData.TalentRank then
			if bData.TalentRank == 1 then
				if modType == "Crit" then
					critPerc = critPerc + bData.Value * BuffTalentRank
				elseif not modType then
					dmgM = dmgM * ( 1 + bData.Value * BuffTalentRank )
				end
			end
		elseif modType == "BaseDamage" then
			bDmgM = bDmgM + bData.Value
		elseif modType == "SpellDamage" then
			spellDmgM = spellDmgM + bData.Value
		elseif modType == "Crit" then
			critPerc = critPerc + bData.Value			
		elseif not modType then
			dmgM = dmgM * ( 1 + bData.Value )
		end	
	end
		
	
	
	--Start calculating Buffs/Debuffs that affects damage not provided by blizzard API.
	local socStun = false			--Stun for SoC
	local afflCount = 0			--Count (affliction) effects for life drain	
	
	if DrDamage.Debuffs and not baseSpell.NoDebuffs then
	
		local incinerateBonus = false 		--To make sure Immolate bonus to Incinerate is only added once
		local iceLanceBonus = false		--To make sure Ice Lance get's bonus only once

		for buffName, apps, _, texture, rank in SEA:DebuffIter("target") do
		
			if BS:HasReverseTranslation( buffName ) then
				buffName = BS:GetReverseTranslation( buffName )
			end
	
			if DrDamage.Debuffs[buffName] then
			
				local buffData = DrDamage.Debuffs[buffName]
				local modType = buffData.ModType				
				
				if SSRank > 0 then
					if buffData.Affliction then
						afflCount = afflCount + 1
					end
				end
				
				if spellName == "Seal of Command" and not socStun and buffData.Stun then
					socStun = true
				end
							
				if not healingSpell 
					and ( DrDamage:MatchData( buffData.School, spellSchool ) or not buffData.School and not buffData.Spell )
				or healingSpell and buffData.School == "Healing"
				or DrDamage:MatchData( buffData.Spell, spellName )
					and not buffData.Stun
					and not buffData.Affliction
				then
					if modType == "Special" then
						if buffName == "Immolate" and not incinerateBonus then
							minDam = minDam + spell.incinerateMin
							maxDam = maxDam + spell.incinerateMax
							incinerateBonus = true
						elseif ( buffName == "Frost Nova" or buffName == "Frostbite" ) and not iceLanceBonus then
							dmgM = dmgM * 3
							iceLanceBonus = true
						end
					else
						BuffCalc( buffName, buffData, apps, texture, nil )
					end
				end
			end
		end
	end
		
	if DrDamage.DamageBuffs and not baseSpell.NoBuffs then
	
		for buffName, index, apps, _, rank in SEA:BuffIter( "player" ) do
		
			if BS:HasReverseTranslation( buffName ) then
				buffName = BS:GetReverseTranslation( buffName )
			end		
		
			if DrDamage.DamageBuffs[buffName] then
						
				local buffData = DrDamage.DamageBuffs[buffName]
				
				if not healingSpell 
					and ( DrDamage:MatchData( buffData.School, spellSchool ) or ( not buffData.School and not buffData.Spell ))
				or healingSpell 
					and buffData.School == "Healing"
				or buffData.School == "All"
				or DrDamage:MatchData( buffData.Spell, spellName ) then
				
					if buffData.ModType == "Special" then
						if buffName == "Fel Armor" then
							LeechBonus = LeechBonus * 1.2
						elseif buffName == "Master Demonologist" then
							if UnitCreatureFamily( "pet" ) then
								local petFamily = UnitCreatureFamily("pet")
								
								if petFamily == "Succubus" and BuffTalentRank > 0 then
									dmgM = dmgM * ( 1 + 0.02 * BuffTalentRank )				
								elseif petFamily == "Felguard" then
									dmgM = dmgM * ( 1 + 0.01 * BuffTalentRank )
								end
							end
						elseif buffName == "Touch of Shadow" then
							GT:SetUnitBuff( "player", index )
							local _, _, amount = GT:Find( "(%d+)%%" )
							
							if amount == "10" then
								dmgM = dmgM * 1.1
							else
								dmgM = dmgM * 1.15
							end	
						elseif buffName == "Heroism" or buffName == "Bloodlust" then
							castTime = castTime * 0.65
						elseif buffName == "Totem of Wrath" then
							hitPerc = hitPerc + 3
							critPerc = critPerc + 3
						end
					else
						BuffCalc( buffName, buffData, apps, nil, index )
					end				
				end
			end		
		end
	end
	
	if DrDamage.DamageDebuffs and not baseSpell.NoBuffs then
	
		for buffName, apps, _, texture, rank in SEA:DebuffIter( "player" ) do
		
			if BS:HasReverseTranslation( buffName ) then
				buffName = BS:GetReverseTranslation( buffName )
			end		
		
			if DrDamage.DamageDebuffs[buffName] then
						
				local buffData = DrDamage.DamageDebuffs[buffName]
				
				if not healingSpell 
					and ( DrDamage:MatchData( buffData.School, spellSchool ) or ( not buffData.School and not buffData.Spell ))
				or DrDamage:MatchData( buffData.Spell, spellName ) then	
				
					if buffData.ModType == "Special" then
						if buffName == "Arcane Blast" then
							castTime = castTime - (1/3) * apps
						end
					else
						BuffCalc( buffName, buffData, apps, texture, nil )
					end
				end
			end
		end
	end
		
	if healingSpell and DrDamage.HealingBuffs then
		for buffName, index, apps, _, rank in SEA:BuffIter("target") do
		
			if BS:HasReverseTranslation( buffName ) then
				buffName = BS:GetReverseTranslation( buffName )
			end		
	
			if DrDamage.HealingBuffs[buffName] then
			
				local buffData = DrDamage.HealingBuffs[buffName]
				
				if not buffData.Spell or buffData.Spell and buffData.Spell == spellName then
					BuffCalc( buffName, buffData, apps, nil, index )
				end
			end
		end
	end
	
	
	--Calculate in the most important set bonuses
	
	if type( DrDamage.SetBonuses ) == "function" then
		eDuration, finalMod, dmgM, castTime, chainFactor = DrDamage:SetBonuses( spellName, eDuration, finalMod, dmgM, castTime, chainFactor )
	end
	
	--Modified factors?
	
	--Split between bonus to dot and nuke
	if baseSpell.hybridFactor then
		spellDmgM = spellDmgM * baseSpell.hybridFactor
	end
	
	--Factor that applies to all ranks of the spell
	if baseSpell.sFactor then
		spellDmgM = spellDmgM * baseSpell.sFactor
	end
	
	--Factor that applies to current rank of the spell
	if spell.sFactor then
		spellDmgM = spellDmgM * spell.sFactor
	end
	
	
	--Factor for downranking
	local lowRankFactor = 1
	
	if spell.maxLevel and spell.maxLevel < 68 then
		
		local playerLevel = UnitLevel("player")
		
			local downRankMod = 0
			
			if spell.Downrank then
				downRankMod = spell.Downrank
			elseif baseSpell.Downrank then
				downRankMod = baseSpell.Downrank
			end
	
			if healingSpell then
				lowRankFactor = ( spell.maxLevel + downRankMod + 6 ) / playerLevel
			else
				lowRankFactor = ( spell.maxLevel + downRankMod + 4 ) / playerLevel
			end
		
			if lowRankFactor < 1 then
				spellDmgM = spellDmgM * lowRankFactor
			else
				lowRankFactor = 1
			end
	end
	
	--Factor for spells under level 20
	local lowLevelFactor = 1
	
	if spell.spellLevel and spell.spellLevel < 20 then
	
		lowLevelFactor = 1 - ((20 - spell.spellLevel) * 0.0375)
		
		if lowLevelFactor < 1 and lowLevelFactor > 0 then
			spellDmgM = spellDmgM * lowLevelFactor
		else
			lowLevelFactor = 1
		end
	end
	
	--Adding to spells base damage after levelups:
	if baseSpell.BaseIncrease and spell.spellLevel then
	
		local playerLevel = UnitLevel("player")
	
		if playerLevel > spell.spellLevel then
				
			if spell.maxLevel and playerLevel >= spell.maxLevel then
				minDam = minDam + spell[3]
				maxDam = maxDam + spell[4]
				
			elseif baseSpell.NextRank and playerLevel >= ( spell.spellLevel + baseSpell.NextRank - 1 ) then
				minDam = minDam + spell[3]
				maxDam = maxDam + spell[4]			
			else
				local diff = 1
				
				if baseSpell.NextRank then
					diff = baseSpell.NextRank - 1
				elseif spell.maxLevel then
					diff = spell.maxLevel - spell.spellLevel
				end
				
				minDam = minDam + ( playerLevel - spell.spellLevel ) * ( spell[3] / diff )
				maxDam = maxDam + ( playerLevel - spell.spellLevel ) * ( spell[4] / diff )
			end
		end
		
		minDam = DrDamage:Round( minDam )
		maxDam = DrDamage:Round( maxDam )
	end
	
	--Special cases:
	
	--Seal of Command:
	if socStun then
		minDam = minDam * 2
		maxDam = maxDam * 2
	end
	
	--Soul siphon
	if SSRank > 0 then

		local SSBonus = 0

		if SSRank == 1 then
			SSBonus = afflCount * 0.02

			if SSBonus > 0.24 then
				SSBonus = 0.24
			end
		elseif SSRank == 2 then
			SSBonus = afflCount * 0.05

			if SSBonus > 0.60 then
				SSBonus = 0.60
			end
		end

		dmgM = dmgM * ( 1 + SSBonus )	
	end
	
	function DmgCalc( next, hitCalc )
	
		--Initialize variables that are calculated
		local calcAvgDmgCrit = 0		--Calculated final damage with crit
		local calcAvgDmg = 0			--Calculated final damage without crit
		local calcMinDmg = 0			--Calculated dmg range: MIN
		local calcMaxDmg = 0			--Calculated dmg range: MAX
		local calcMinCrit = 0
		local calcMaxCrit = 0		
		local calcDotDmg = 0			--Calculated DOT portion (if applicable)
		local igniteDotDmg = 0			--Calculated Ignite DOT Portion
		local avgIgnite	= 0			--Calculated Average Ignite
		local dispSpellDmgM = spellDmgM		--Calculated spelldmg M
		local perHit = 0
	
		
		if sHits then
			calcMinDmg = dmgM * ( bDmgM * minDam * sHits + (  spellDmg * spellDmgM ) )
			calcMaxDmg = dmgM * ( bDmgM * maxDam * sHits + (  spellDmg * spellDmgM ) )
		elseif spell.hybridDotDmg then
			calcMinDmg = ( dmgM + bNukeDmg ) * ( bDmgM * minDam + ( spellDmg * spellDmgM ) )
			calcMaxDmg = ( dmgM + bNukeDmg ) * ( bDmgM * maxDam + ( spellDmg * spellDmgM ) )
		else
			calcMinDmg = dmgM * ( bDmgM * minDam + ( spellDmg * spellDmgM ) )
			calcMaxDmg = dmgM * ( bDmgM * maxDam + ( spellDmg * spellDmgM ) )
		end
				
		if finalMod ~= 0 then
		
			calcMinDmg = calcMinDmg + finalMod
			calcMaxDmg = calcMaxDmg + finalMod
			
			if calcMinDmg < 0 then
				calcMinDmg = 0
			end
			
			if calcMaxDmg < 0 then
				calcMaxDmg = 0
			end	
		end				
		
		calcAvgDmg = (calcMinDmg + calcMaxDmg) / 2
				
		--This is for Imp SW:P and others with extended duration
		if baseSpell.eDot and baseSpell.eDuration and eDuration > baseSpell.eDuration then
			local bonusDotDamage = ( calcAvgDmg / baseSpell.eDuration ) * ( eDuration - baseSpell.eDuration )
			calcAvgDmg = calcAvgDmg + bonusDotDamage
		end		
		
		--Crit calculation:
		local critBonus = 0
		
		if critPerc > 100 then
			critPerc = 100
		end
		
		if baseSpell.canCrit then
			calcMinCrit = calcMinDmg + calcMinDmg * critM
			calcMaxCrit = calcMaxDmg + calcMaxDmg * critM
			critBonus = ( critPerc / 100 ) * calcAvgDmg * critM
			calcAvgDmgCrit = calcAvgDmg + critBonus
		else
			calcAvgDmgCrit = calcAvgDmg
			critPerc = 0;
		end

		--Hit calculation:
		if hitPerc > 99 then
			hitPerc = 99
		end
		
		if not healingSpell and not baseSpell.Unresistable then
			if DrDamage.db.profile.HitCalc or hitCalc then
				calcAvgDmgCrit = ( calcAvgDmgCrit - critBonus ) * ( hitPerc / 100 ) + critBonus
			end
		end		
		
		--Hybrid spells
		if spell.hybridDotDmg then
	
			local hybridFactor = 0
			local sFactor = 1
			
				if baseSpell.hybridFactor then
					hybridFactor = baseSpell.hybridFactor
				end
				
				if baseSpell.sFactor then
					sFactor = baseSpell.sFactor * sFactor
				end
				
				if spell.sFactor then
					sFactor = spell.sFactor * sFactor
				end
				
				if baseSpell.hybridDotFactor then
					sFactor = baseSpell.hybridDotFactor * sFactor
				end
			
			local baseDuration = spell.eDuration or baseSpell.eDuration
		
			calcDotDmg = dmgM * ( spell.hybridDotDmg + ( baseDuration / 15 + dotSpellDmgM ) * ( 1 - hybridFactor ) * sFactor * lowRankFactor * lowLevelFactor * spellDmg )
			dispSpellDmgM = dispSpellDmgM +  ( baseDuration / 15 + dotSpellDmgM ) * ( 1 - hybridFactor ) * sFactor * lowRankFactor * lowLevelFactor
		
			if eDuration > baseDuration then
				local bonusDotDamage = ( calcDotDmg / baseDuration ) * ( eDuration - baseDuration )
				calcDotDmg = calcDotDmg + bonusDotDamage
			end
		end
		
			
		
		--For spells that can ignite
		if baseSpell.Ignite then
			igniteDotDmg = igniteM * (( calcMinCrit + calcMaxCrit ) / 2 )
			avgIgnite = ( critPerc / 100 ) * igniteDotDmg
		end
		
		if spell.extraDotDmg then
		
			local xFactor = 0
			
			if baseSpell.extraDotF then
				xFactor = baseSpell.extraDotF
			end
			
			calcDotDmg = calcDotDmg + dmgM * ( spell.extraDotDmg + xFactor * lowRankFactor * lowLevelFactor * spellDmg ) --TODO: Verify factor
			dispSpellDmgM = dispSpellDmgM + xFactor * lowRankFactor * lowLevelFactor
		end
		
		--For chain effects (chain lightning, chain heal)
		local chainEffect = 0
		
		if chainFactor > 0 then
			chainEffect = calcAvgDmgCrit * ( chainFactor + math.pow( chainFactor, 2 ) + math.pow( chainFactor, 3 ) )
		end
					
		--Begin DPS calculations
		
		local calcDPS = 0
		local calcDPT = 0
		local calcDPSCooldown = 0
		
		if castTime < 1.5 then
			castTime = 1.5
		end
	
		if baseSpell.eDot then
			calcDPS = calcAvgDmgCrit / eDuration
		else
			if baseSpell.Ignite then
			
				calcDPS = calcAvgDmgCrit / castTime + avgIgnite / 4
				
				if spell.hybridDotDmg or spell.extraDotDmg then
					calcDPS = calcDPS + calcDotDmg / eDuration
				end
				
				calcAvgDmgCrit = calcAvgDmgCrit + avgIgnite + calcDotDmg
				
			elseif spell.hybridDotDmg or spell.extraDotDmg then
				calcDPS = calcAvgDmgCrit / castTime + calcDotDmg / eDuration
				calcAvgDmgCrit = calcAvgDmgCrit + calcDotDmg
			else
				calcDPS = calcAvgDmgCrit / castTime
			end
		end
		
		calcDPT = calcAvgDmgCrit / castTime
		
		if baseSpell.Cooldown then
			calcDPSCooldown = calcAvgDmgCrit / cooldown
		end
						
		if not next then
		
			local DPS = 			DrDamage:Round( calcDPS, 1 )
			local DPSC = 			DrDamage:Round( calcDPT, 1 )
			local Avg =			DrDamage:Round( calcAvgDmg )
			local MinDmg = 			math.floor( calcMinDmg )
			local MaxDmg = 			math.ceil( calcMaxDmg )			
			CalculationResults.MinCrit =	math.floor( calcMinCrit )
			CalculationResults.MaxCrit = 	math.ceil( calcMaxCrit )
			CalculationResults.AvgCrit = 	DrDamage:Round( ( calcMinCrit + calcMaxCrit ) / 2 )
			CalculationResults.CritRate = 	DrDamage:Round( critPerc, 2 )
			CalculationResults.HitRate = 	DrDamage:Round( hitPerc, 2 )
			CalculationResults.DotDmg = 	DrDamage:Round( calcDotDmg )
			CalculationResults.SpellDmg = 	DrDamage:Round( spellDmg )
			CalculationResults.SpellDmgM = 	DrDamage:Round( dispSpellDmgM, 3 )
			CalculationResults.DmgM = 	DrDamage:Round( dmgM, 3 )
			CalculationResults.IgniteDmg = 	DrDamage:Round( igniteDotDmg )
			CalculationResults.DPSCD = 	DrDamage:Round( calcDPSCooldown, 1 )
			CalculationResults.ChainDmg = 	DrDamage:Round( chainEffect )
			CalculationResults.CastTime = 	castTime
			CalculationResults.Cooldown = 	cooldown
			CalculationResults.ManaRegen = 	DrDamage:Round( manaRegenCasting, 1 )
			
			if englishClass == "PALADIN" then
				if illumination > 0 then
					CalculationResults.Illumination = illumination
				else
					CalculationResults.Illumination = 0
				end
			end
			
			local sTicks = baseSpell.sTicks or spell.sTicks 
		
			if sHits then
				perHit = calcAvgDmg / sHits
				CalculationResults.Hits = sHits
			elseif sTicks then
				local ticks
				if castTime > eDuration then
					ticks = castTime / sTicks
				else
					ticks = eDuration / sTicks
				end
				
				CalculationResults.Hits = ticks
				
				if spell.hybridDotDmg or spell.extraDotDmg then
					perHit = calcDotDmg / ticks
				else
					perHit = calcAvgDmg / ticks
				end
			end		
		
			if perHit > 0 then
				CalculationResults.PerHit = DrDamage:Round( perHit, 1 )
				
				if baseSpell.Leech then
					CalculationResults.PerHitHeal =  DrDamage:Round( perHit * LeechBonus, 1 )
				end
			end
			
			if spellName == "Seed of Corruption" then
				ticks = math.ceil( 1044 / perHit )
				DPS = DrDamage:Round( calcDPS -  ( calcDotDmg / eDuration ) +  ( perHit / 3 ), 1 )
				calcAvgDmgCrit = calcAvgDmgCrit - calcDotDmg + ticks * perHit
				DPSC = DrDamage:Round( calcAvgDmgCrit / castTime, 1 )
				CalculationResults.Hits = ticks
				CalculationResults.DotDmg = ticks * perHit
			end
				
			
			if baseSpell.Leech then
				CalculationResults.AvgLeech = DrDamage:Round( calcAvgDmgCrit * LeechBonus, 1 )
			end
			
			return calcAvgDmgCrit, DPS, DPSC, Avg, MinDmg, MaxDmg

		end
		
		--Special cases:
		if spellName == "Seed of Corruption" then
			calcAvgDmgCrit = calcAvgDmgCrit - calcDotDmg + CalculationResults.Hits * CalculationResults.PerHit
		end		
		
		return calcAvgDmgCrit
	end
	
	local returnAvgTotal, returnDPS, returnDPSC, returnAvg, returnMinDmg, returnMaxDmg = DmgCalc()
	
	local returnNextCrit = 0
	local returnNextTenDmg = 0
	local returnNextHit = 0
	
	if DrDamage.db.profile.Next then
	
		if baseSpell.canCrit then
			critPerc = critPerc + 1
			returnNextCrit = DrDamage:Round( DmgCalc( true ) - returnAvgTotal, 1 )
			critPerc = critPerc - 1	
		end
	
		spellDmg = spellDmg + 10
		returnNextTenDmg = DrDamage:Round( DmgCalc( true ) - returnAvgTotal, 1 )
		spellDmg = spellDmg - 10
	
		if not healingSpell then
			local temp = DmgCalc( true, true )
			hitPerc = hitPerc + 1
			returnNextHit = DrDamage:Round( DmgCalc( true, true ) - temp, 1 )
			hitPerc = hitPerc - 1
		end
	end
	
	returnAvgTotal = DrDamage:Round( returnAvgTotal )
	
	CalculationResults.NextCrit = returnNextCrit
	CalculationResults.NextTenDmg = returnNextTenDmg
	CalculationResults.NextHit = returnNextHit
	
	return healingSpell, returnAvgTotal, returnDPS, returnDPSC, returnAvg, returnMinDmg, returnMaxDmg, CalculationResults
end

function DrDamage:Round(x, y)

	if type( x ) ~= "number" then
		return x
	end

	local s
	
	if y then 
		x = x * math.pow(10, y) 
	end
	
	if (x > 0) then
		s = math.floor(x+0.5);
  	else
    		s = math.ceil(x-0.5);
  	end
  	
  	if y then 
  		s = s / math.pow(10, y) 
  	end
  	
  	return s
end

